unit HuffFuncs;

//   Huffman Compression Algorithm
//   Based on source (public domain?) found at:
//
//   http://www.howtodothings.com/showarticle.asp?article=313
//   (C) Mr Vimil Saju (vimilsaju@yahoo.com)
//
//   Modified July-August 2002 (C) Restin (restin@rambler.ru)
//
//   Added:
//       TBitWriter for comfort bit streaming
//       Word size of huffman codes
//       Output for reigncom devices
//	 Some logging subroutines (not used now)

interface

uses classes,sysutils,windows;

type
 Phuffinfo=^Thuffinfo;
 THuffInfo=record
  left:phuffinfo;
  right:phuffinfo;
  code: array[0..256] of word;
  codecount:integer;
  huff,char:integer;
  freq:integer;
  ticked:boolean;
 end;
 THuffcode=record
  char:integer;
  used:boolean;
  code:array[0..255] of byte;
  codelength:integer;
 end;
 PhuffCode=^THuffCode;

        TReignComHDR = record
                       id:array[0..15] of char;
                       mnf:array[0..15] of char;
                       size:integer;
                       checksum:integer;
                       end;

          TBitWriter = class
                       constructor Create(stream:TStream);
                       procedure WriteBit(value:integer);
                       procedure Finalize;
                       function GetCheckSum:integer;
                       private
                       FStream: TStream;
                       FOutBuffer: array[0..32767] of byte;
                       fk,fbit,fchecksum:integer;
                       ftmpcode:byte;
                       end;

type
 EError=class(Exception);
procedure Initialize(id,mnf:string);
function SetInputfile(afilename:string):boolean;
function SetOutputfile(afilename:string):boolean;
procedure Compress;
procedure Decompress;
procedure Finalize;


var
 huffcodes: array[0..256] of Thuffcode;

implementation

procedure GetDistribution;forward;
procedure InitList;forward;
procedure BuildTree;forward;
procedure GetCodes;forward;
procedure SaveTable;forward;
procedure RetrieveTable; forward;
procedure ReconstructTree;forward;
procedure WriteCompressedFile;forward;
procedure WriteUncompressedFile;forward;

var
 Charlist:array[0..513] of THuffinfo;
 ifile,ofile:Tfilestream;
 infile,outfile:string;
 InBuffer,OutBuffer:array[0..40000] of byte;
 ufilesize,cfilesize:integer;
 rootnode:Phuffinfo;
 table: array[0..1495] of byte;
 tabsize:integer;
 endbits:byte;
 header:TReignComHDR;

// TBitWriter // TBitWriter // TBitWriter // TBitWriter // TBitWriter //

constructor TBitWriter.Create(Stream:TStream);
begin
     FStream:=Stream;
     fk:=0;
     fbit:=0;
     ftmpcode:=0;
     fchecksum:=0;
     fillchar(FOutBuffer,sizeof(FOutBuffer),0);
end;

procedure TBitWriter.WriteBit(value:integer);
begin
      ftmpcode:=(ftmpcode shl 1) or (value and 1);
      inc(fbit);
      if fbit=8 then
       begin
        outbuffer[fk]:=ftmpcode;
        inc(fchecksum,ftmpcode);
        inc(fk);
        fbit:=0;
        ftmpcode:=0;
       end;
      if fk=32768 then
       begin
        FStream.write(outbuffer,32768);
        fk:=0;
       end;
end;

procedure TBitWriter.Finalize;
begin
 if fbit>0 then
  begin
   ftmpcode:=ftmpcode shl (8-fbit);
   outbuffer[fk]:=ftmpcode;
   inc(fchecksum,ftmpcode);
   fk:=fk+1;
  end;
 if fk>0 then
  FStream.write(outbuffer,fk);
end;

function TBitWriter.GetCheckSum:integer;
begin
     result:=fchecksum;
end;

// TBitWriter // TBitWriter // TBitWriter // TBitWriter // TBitWriter //

procedure Initialize(id,mnf:string);
var i:integer;
begin
 zeromemory(@charlist,sizeof(charlist));
 zeromemory(@table,sizeof(table));
 zeromemory(@huffcodes,sizeof(huffcodes));
 infile:='';
 outfile:='';
 endbits:=0;
 ufilesize:=0;
 cfilesize:=0;
 rootnode:=nil;

     fillchar(header,sizeof(header),0);

     for i:=1 to sizeof(header.id)-1 do
     begin
       if i>length(id) then break;
       header.id[i-1]:=id[i];
     end;

     for i:=1 to sizeof(header.mnf)-1 do
     begin
       if i>length(mnf) then break;
       header.mnf[i-1]:=mnf[i];
     end;
end;

procedure Finalize;
begin
 ifile.free;
 ofile.free;
end;

function SetInputfile(afilename:string):boolean;
begin
 result:=fileexists(afilename);
 if result=true then
  begin
   ifile:=Tfilestream.create(afilename,fmOpenRead or fmShareDenyNone);
   infile:=afilename;
   ufilesize:=GetFileSize(ifile.Handle,nil);
  end
 else
  raise EError.Create('Invalid Filename');

end;

procedure InitList;
var
 i:integer;
begin
 zeromemory(@charlist,sizeof(charlist));
 zeromemory(@huffcodes,sizeof(huffcodes));
  for i:=0 to 256 do
  begin
   charlist[i].code[charlist[i].codecount]:=i;
   inc(charlist[i].codecount);
  end;
end;

function SetOutputfile(afilename:string):boolean;
begin
 ofile:=Tfilestream.create(afilename,fmCreate or fmShareDenyNone);
 outfile:=afilename;
 result:=true;
end;

procedure GetDistribution;
var
 i:integer;
 bufcount:integer;
begin
 bufcount:=ifile.read(inbuffer,32768);
 repeat
  for i:=0 to bufcount-1 do
   charlist[inbuffer[i]].freq:=charlist[inbuffer[i]].freq+1;
  bufcount:=ifile.read(inbuffer,32768);
 until bufcount=0;
 charlist[256].freq:=1;
end;

procedure dumpcharlist(fl:string);
var f:textfile;
    i,j:integer;
begin
 assignfile(F,'/var/log/love/sees/no/colour.log'+fl);
 Rewrite(F);
 for i:=0 to 513 do
 begin
   write(f,i,' - ',charlist[i].freq);
   if charlist[i].left<>nil then write(f,',',charlist[i].left.char);
   if charlist[i].right<>nil then write(f,',',charlist[i].right.char);
   write(f,' : ');
   for j:=0 to charlist[i].codecount-1 do
    write(f,charlist[i].code[j],';');
   writeln(f);
 end;
 close(f);
end;

procedure BuildTree;
var
 i,cnt,tmp:integer;
 pinfo1,pinfo2:Phuffinfo;
begin
 pinfo1:=nil;
 pinfo2:=nil;
 cnt:=256;
 while true do
  begin
   tmp:=maxint;
   for i:=0 to cnt do
    begin
     if (charlist[i].freq<tmp) and (charlist[i].freq > 0) and(charlist[i].ticked=false) then
      begin
       pinfo1:=@charlist[i];
       tmp:=pinfo1.freq;
      end;
    end;
   if pinfo1=nil then
    break;
   pinfo1.ticked:=true;
   tmp:=maxint;
   for i:=0 to cnt do
    begin
     if (charlist[i].freq<tmp) and (charlist[i].freq > 0) and (charlist[i].ticked=false) then
      begin
       pinfo2:=@charlist[i];
       tmp:=pinfo2.freq;
      end;
    end;
   if pinfo2=nil then
    break;
   pinfo2.ticked:=true;
   inc(cnt);
   charlist[cnt].freq:=pinfo1.freq+pinfo2.freq;

   move(pinfo1.code,charlist[cnt].code,pinfo1.codecount*sizeof(pinfo1.code[0]));
   move(pinfo2.code,charlist[cnt].code[pinfo1.codecount],pinfo2.codecount*sizeof(pinfo1.code[0]));

   charlist[cnt].codecount:=pinfo1.codecount+pinfo2.codecount;
   charlist[cnt].left:=pinfo2;
   charlist[cnt].right:=pinfo1;
   pinfo1:=nil;
   pinfo2:=nil;
  end;
 rootnode:=@charlist[cnt];
end;

procedure GetCodes;
var
 i,j:integer;
 tmpnode:phuffinfo;
 flag:integer;
begin
 for i:=0 to 256 do
  begin
   flag:=-1;
   tmpnode:=rootnode;
   while tmpnode.left<>nil do
    begin
     for j:=0 to tmpnode.left.codecount-1 do
      if tmpnode.left.code[j]= i then
       begin
        flag:=0;
        tmpnode:=tmpnode.left;
        break;
       end;
      if flag=-1 then
       begin
        for j:=0 to tmpnode.right.codecount-1 do
         if tmpnode.right.code[j]= i then
          begin
           flag:=1;
           tmpnode:=tmpnode.right;
           break;
          end;
       end;
     if flag=-1 then
      break;
     huffcodes[i].used:=true;
     huffcodes[i].code[huffcodes[i].codelength]:=flag;
     huffcodes[i].codelength:=huffcodes[i].codelength+1;
     huffcodes[i].char:=i;
     flag:=-1;
   end;
  end;
end;

procedure writeheader(Stream:TStream;sz,cs:integer);
begin
    header.size:=sz;
    header.checksum:=cs;
    stream.write(header,sizeof(header));
end;

procedure writeencodedtable(writer:TBitWriter);
var i,j,v:integer;
begin
     writer.writebit(1); // 256 bit flag

     for i:=0 to 255 do
      // write node presense flag
      writer.writebit(integer(huffcodes[i].used));

     for i:=0 to 256 do
      // note. ENDOFSTREAM (code 256) is always present/used
      // however ENDOFSTREAM presense is not saved
      if huffcodes[i].used then
      begin
       v:=huffcodes[i].codelength-1;
       writer.writebit(0); // 5-bit encoding flag
       for j:=4 downto 0 do // write five-bit encoding
       begin
        writer.writebit((v and (1 shl j)) shr j)
       end;
       for j:=0 to v do // now write huffman code
       begin
        writer.writebit(huffcodes[i].code[j]);
       end;
      end;
end;


procedure WriteCompressedFile;
var
 i,j:integer;
 bufcount:integer;
 writer:tbitwriter;
begin
 ofile.seek(0,sofrombeginning);
 ifile.seek(0,soFromBeginning);
 writer:=TBitWriter.Create(ofile);
 writeheader(ofile,0,0);
 writeencodedtable(writer);
 bufcount:=ifile.read(inbuffer,32768);
 repeat
  for i:=0 to bufcount-1 do
   begin
    for j:=0 to huffcodes[inbuffer[i]].codelength-1 do
      writer.writebit(huffcodes[inbuffer[i]].code[j]);
   end;
  bufcount:=ifile.read(inbuffer,32768);
 until bufcount=0;

 for j:=0 to huffcodes[256].codelength-1 do
   writer.writebit(huffcodes[256].code[j]);

 writer.Finalize;
 ofile.seek(0,0);
 writeheader(ofile,ofile.size-sizeof(TReignComHDR),writer.GetCheckSum);
end;


procedure Compress;
begin
 Initlist;
 GetDistribution;
 BuildTree;
 GetCodes;
 WriteCompressedFile;
end;

procedure SaveTable;
var
 F:textfile;
 i,j:integer;
begin
 assignfile(F,'/var/log/and/the/days/go/by.log');
 Rewrite(F);
 for i:=0 to 256 do
  begin
   if huffcodes[i].used=false then
   begin
    writeln(f,i,' - not used');
    continue;
   end;
   write(f,i,' - ',huffcodes[i].codelength,':');
   for j:=0 to huffcodes[i].codelength-1 do
    begin
     write(f,huffcodes[i].code[j]);
     if j<huffcodes[i].codelength-1 then write(f,',');
    end;
   writeln(f);
  end;
 CloseFile(F)
end;

procedure RetrieveTable;
var
 j,k,l:integer;
 index:integer;
 length,cnt:integer;
 tmpcode:byte;
begin
 ifile.seek(0,soFromBeginning);
 ifile.read(endbits,1);
 if endbits=0 then
  endbits:=8;
 ifile.Read(table,sizeof(table));
 tabsize:=pinteger(@table[0])^;
 k:=sizeof(tabsize);
 while k<tabsize+sizeof(tabsize) do
  begin
   j:=0;
   index:=table[k];
   k:=k+1;
   length:=table[k];
   k:=k+1;
   huffcodes[index].used:=true;
   huffcodes[index].char:=index;
   huffcodes[index].codelength:=length;
   while j<=length-1 do
    begin
     tmpcode:=table[k];
     k:=k+1;
     if (length-j)>8 then
      cnt:=8
     else
      cnt:=length-j;
     for l:=1 to cnt do
      begin
       huffcodes[index].code[l+j-1]:=(tmpcode shr (8-l)) and 1;
      end;
     j:=j+cnt;
    end;
  end;
end;

procedure ReconstructTree;
var
 rinfo:phuffinfo;
 i,j,k:integer;
begin
 k:=0;
 zeromemory(@charlist,sizeof(charlist));
 for i:=0 to 255 do
  begin
   rinfo:=@charlist[511];
   for j:=0 to huffcodes[i].codelength-1 do
    begin
     charlist[k].huff:=huffcodes[i].code[j];
     if huffcodes[i].code[j]= 0 then
      begin
       if rinfo.left=nil then
        begin
         rinfo.left:=@charlist[k];
         k:=k+1;
        end;
       rinfo:=rinfo.left;
      end
     else
      begin
       if rinfo.right=nil then
        begin
         rinfo.right:=@charlist[k];
         k:=k+1;
        end;
       rinfo:=rinfo.right;
      end;
     if j=huffcodes[i].codelength-1 then
      rinfo.char:=huffcodes[i].char;
    end;
  end;
end;

procedure WriteUncompressedFile;
var
 bufcount,i,j,l:integer;
 tmpbit,tmpbyte:byte;
 tmpnode:phuffinfo;
begin
 i:=0;j:=1;l:=0;
 tmpnode:=@charlist[511];
 ifile.seek(tabsize+sizeof(tabsize)+1,soFromBeginning);
 ofile.seek(0,soFromBeginning);
 bufcount:=ifile.read(inbuffer,32768);
 tmpbyte:=inbuffer[0];
 i:=i+1;
 repeat
  while true do
   begin
    while (tmpnode.left<>nil) do
     begin
      if j>8 then
       begin
        tmpbyte:=inbuffer[i];
        i:=i+1;
        if i=bufcount then
         begin
          bufcount:=ifile.read(inbuffer,32768);
          i:=0;
         end;
        j:=1;
       end;
      tmpbit:=(tmpbyte shr (8-j)) and 1;
      if tmpnode.left.huff=tmpbit then
       begin
        tmpnode:=tmpnode.left;
        j:=j+1;
       end
      else
       begin
        tmpnode:=tmpnode.right;
        j:=j+1;
       end;
     end;
    OutBuffer[l]:=tmpnode.char;
    l:=l+1;
    tmpnode:=@charlist[511];
    if l=32768 then
     begin
      ofile.write(outbuffer,32768);
      l:=0;
     end;
    if (bufcount=0) and (i=0) and (j>endbits) then
     break;
   end;
 until bufcount=0;
 if l>0 then
  ofile.write(outbuffer,l);
end;

procedure Decompress;
begin
 RetrieveTable;
 ReConstructTree;
 WriteUncompressedfile;
end;

end.
